@@ -1,4 +1,8 @@
+from functools import wraps
+import json
 import logging
+from couchdbkit.exceptions import ResourceNotFound
+from psycopg2._psycopg import InterfaceError
 import pytz
 from datetime import datetime
 import hashlib
@@ -9,21 +13,28 @@ import time
 
 from django.core.mail import send_mail
 from requests import ConnectionError
+import requests
 import simplejson
 import rawes
 from django.conf import settings
+import sys
 
 from dimagi.utils.decorators.memoized import memoized
+from dimagi.utils.couch import LockManager
+from pillow_retry.models import PillowError
 from pillowtop.couchdb import CachedCouchDB
 from .utils import import_settings
 
 from couchdbkit.changes import ChangesStream
+from django import db
+from dateutil import parser
 
 
 pillow_logging = logging.getLogger("pillowtop")
 pillow_logging.setLevel(logging.INFO)
 
 CHECKPOINT_FREQUENCY = 100
+CHECKPOINT_MIN_WAIT = 300
 WAIT_HEARTBEAT = 10000
 CHANGES_TIMEOUT = 60000
 RETRY_INTERVAL = 2  # seconds, exponentially increasing
@@ -84,6 +95,10 @@ class PillowtopNetworkError(Exception):
     pass
 
 
+class PillowtopCheckpointReset(Exception):
+    pass
+
+
 def ms_from_timedelta(td):
     """
     Given a timedelta object, returns a float representing milliseconds
@@ -91,23 +106,37 @@ def ms_from_timedelta(td):
     return (td.seconds * 1000) + (td.microseconds / 1000.0)
 
 
+def lock_manager(obj):
+    if isinstance(obj, LockManager):
+        return obj
+    else:
+        return LockManager(obj, None)
+
+
 class BasicPillow(object):
     couch_filter = None  # string for filter if needed
-    extra_args = {} # filter args if needed
+    extra_args = {}  # filter args if needed
     document_class = None  # couchdbkit Document class
     changes_seen = 0
     couch_db = None
     include_docs = True
+    use_locking = False
+    _current_checkpoint = None
 
     def __init__(self, couch_db=None, document_class=None):
         if document_class:
             self.document_class = document_class
 
-        if couch_db:
-            self.couch_db = couch_db
+        # Explicitly check for None since a couch db class will evaluate to False
+        # if there are no docs in the database.
+        self.couch_db = couch_db if couch_db is not None else (
+            self.couch_db if self.couch_db is not None else (
+                self.document_class.get_db() if self.document_class else None
+            ))
 
-        if not self.couch_db:
-            self.couch_db = self.document_class.get_db() if self.document_class else None
+        if self.use_locking:
+            # document_class must be a CouchDocLockableMixIn
+            assert hasattr(self.document_class, 'get_obj_lock_by_id')
 
         self.settings = import_settings()
 
@@ -116,10 +145,24 @@ class BasicPillow(object):
         Couchdbkit > 0.6.0 changes feed listener handler (api changes after this)
         http://couchdbkit.org/docs/changes.html
         """
-        with ChangesStream(self.couch_db, feed='continuous', heartbeat=True, since=self.since,
-                           filter=self.couch_filter, include_docs=self.include_docs, **self.extra_args) as st:
-            for c in st:
-                self.processor(c)
+        changes_stream = ChangesStream(
+            db=self.couch_db,
+            feed='continuous',
+            heartbeat=True,
+            since=self.since,
+            filter=self.couch_filter,
+            include_docs=self.include_docs,
+            **self.extra_args
+        )
+        while True:
+            try:
+                for change in changes_stream:
+                    if change:
+                        self.processor(change)
+                    else:
+                        self.touch_checkpoint(min_interval=CHECKPOINT_MIN_WAIT)
+            except PillowtopCheckpointReset:
+                self.changes_seen = 0
 
     def run(self):
         """
@@ -145,7 +188,7 @@ class BasicPillow(object):
     def get_checkpoint_doc_name(self):
         return "pillowtop_%s" % self.get_name()
 
-    def get_checkpoint(self):
+    def get_checkpoint(self, verify_unchanged=False):
         doc_name = self.get_checkpoint_doc_name()
 
         if self.couch_db.doc_exist(doc_name):
@@ -168,6 +211,12 @@ class BasicPillow(object):
                 "seq": starting_seq
             }
             self.couch_db.save_doc(checkpoint_doc)
+
+        if verify_unchanged and self._current_checkpoint and checkpoint_doc['seq'] != self._current_checkpoint['seq']:
+            raise PillowtopCheckpointReset()
+
+        self._current_checkpoint = checkpoint_doc
+
         return checkpoint_doc
 
     def reset_checkpoint(self):
@@ -183,7 +232,8 @@ class BasicPillow(object):
         return checkpoint['seq']
 
     def set_checkpoint(self, change):
-        checkpoint = self.get_checkpoint()
+        checkpoint = self.get_checkpoint(verify_unchanged=True)
+
         pillow_logging.info(
             "(%s) setting checkpoint: %s" % (checkpoint['_id'], change['seq'])
         )
@@ -191,6 +241,27 @@ class BasicPillow(object):
         checkpoint['timestamp'] = datetime.now(tz=pytz.UTC).isoformat()
         self.couch_db.save_doc(checkpoint)
 
+    def touch_checkpoint(self, min_interval=0):
+        """
+        Update the checkpoint timestamp without altering the sequence.
+        :param min_interval: minimum interval between timestamp updates
+        """
+        checkpoint = self.get_checkpoint(verify_unchanged=True)
+
+        now = datetime.now(tz=pytz.UTC)
+        previous = self._current_checkpoint.get('timestamp')
+        do_update = True
+        if previous:
+            diff = now - parser.parse(previous).replace(tzinfo=pytz.UTC)
+            do_update = diff.total_seconds() >= min_interval
+
+        if do_update:
+            pillow_logging.info(
+                "(%s) touching checkpoint" % checkpoint['_id']
+            )
+            checkpoint['timestamp'] = now.isoformat()
+            self.couch_db.save_doc(checkpoint)
+
     def get_db_seq(self):
         return self.couch_db.info()['update_seq']
 
@@ -201,16 +272,6 @@ class BasicPillow(object):
         """
         self.processor(simplejson.loads(change))
 
-    def process_change(self, change):
-        try:
-            t = self.change_trigger(change)
-            if t is not None:
-                tr = self.change_transform(t)
-                if tr is not None:
-                    self.change_transport(tr)
-        except Exception, ex:
-            pillow_logging.exception("[%s] Error on change: %s, %s" % (self.get_name(), change['id'], ex))
-
     def processor(self, change, do_set_checkpoint=True):
         """
         Parent processsor for a pillow class - this should not be overridden.
@@ -223,6 +284,36 @@ class BasicPillow(object):
 
         self.process_change(change)
 
+    def process_change(self, change, is_retry_attempt=False):
+        try:
+            with lock_manager(self.change_trigger(change)) as t:
+                if t is not None:
+                    tr = self.change_transform(t)
+                    if tr is not None:
+                        self.change_transport(tr)
+        except Exception, ex:
+            if not is_retry_attempt:
+                try:
+                    # This breaks the module boundary by using a show function defined in commcare-hq
+                    # but it was decided that it wasn't worth the effort to maintain the separation.
+                    meta = self.couch_db.show('domain/domain_date', change['id'])
+                except ResourceNotFound:
+                    # Show function does not exist
+                    meta = None
+                error = PillowError.get_or_create(change, self, change_meta=meta)
+                error.add_attempt(ex, sys.exc_info()[2])
+                error.save()
+                pillow_logging.exception(
+                    "[%s] Error on change: %s, %s. Logged as: %s" % (
+                        self.get_name(),
+                        change['id'],
+                        ex,
+                        error.id
+                    )
+                )
+            else:
+                raise
+
     def change_trigger(self, changes_dict):
         """
         Step one of pillowtop process
@@ -235,12 +326,17 @@ class BasicPillow(object):
         Should return a doc_dict
         """
         if changes_dict.get('deleted', False):
-            #override deleted behavior on consumers that care/deal with deletions
+            # override deleted behavior on consumers that deal with deletions
             return None
-        if 'doc' in changes_dict:
+        id = changes_dict['id']
+        if self.use_locking:
+            lock = self.document_class.get_obj_lock_by_id(id)
+            lock.acquire()
+            return LockManager(self.couch_db.open_doc(id), lock)
+        elif 'doc' in changes_dict:
             return changes_dict['doc']
         else:
-            return self.couch_db.open_doc(changes_dict['id'])
+            return self.couch_db.open_doc(id)
 
     def change_transform(self, doc_dict):
         """
@@ -261,6 +357,7 @@ class BasicPillow(object):
 PYTHONPILLOW_CHUNK_SIZE = 250
 PYTHONPILLOW_CHECKPOINT_FREQUENCY = CHECKPOINT_FREQUENCY * 10
 
+
 class PythonPillow(BasicPillow):
     """
     A pillow that does filtering in python instead of couch.
@@ -281,7 +378,7 @@ class PythonPillow(BasicPillow):
         Use chunk_size = 0 to disable chunking
         """
         super(PythonPillow, self).__init__(document_class=document_class)
-        self.change_queue=[]
+        self.change_queue = []
         self.chunk_size = chunk_size
         self.use_chunking = chunk_size > 0
         self.checkpoint_frequency = checkpoint_frequency
@@ -331,14 +428,124 @@ class PythonPillow(BasicPillow):
         super(PythonPillow, self).run()
 
 
-class ElasticPillow(BasicPillow):
+class BulkPillow(BasicPillow):
+    def bulk_builder(self, changes):
+        """
+        Generator function for bulk changes - note each individual change item goes through the pillowtop pathway individually
+        when loading the bulk item, and short of change_transport, it's identical. It would be slightly more efficient if the couch
+        load could be done in bulk for the actual documents, but it's not quite possible without gutting the existing pillowtop API
+        http://www.elasticsearch.org/guide/reference/api/bulk.html
+        bulk loader follows the following:
+        { "index" : { "_index" : "test", "_type" : "type1", "_id" : "1" } }\n
+        { "field1" : "value1" }\n
+        """
+        for change in changes:
+            try:
+                with lock_manager(self.change_trigger(change)) as t:
+                    if t is not None:
+                        tr = self.change_transform(t)
+                        if tr is not None:
+                            self.change_transport(tr)
+                            yield {
+                                "index": {
+                                    "_index": self.es_index,
+                                    "_type": self.es_type,
+                                    "_id": tr['_id']
+                                }
+                            }
+                            yield tr
+            except Exception, ex:
+                pillow_logging.error(
+                    "[%s] Error on change: %s, %s" % (
+                        self.get_name(),
+                        change['id'],
+                        ex
+                    )
+                )
+
+    def process_bulk(self, changes):
+        self.allow_updates = False
+        self.bulk = True
+        bstart = datetime.utcnow()
+        bulk_payload = '\n'.join(map(simplejson.dumps, self.bulk_builder(changes))) + "\n"
+        pillow_logging.info(
+            "%s,prepare_bulk,%s" % (self.get_name(), str(ms_from_timedelta(datetime.utcnow() - bstart) / 1000.0)))
+        send_start = datetime.utcnow()
+        self.send_bulk(bulk_payload)
+        pillow_logging.info(
+            "%s,send_bulk,%s" % (self.get_name(), str(ms_from_timedelta(datetime.utcnow() - send_start) / 1000.0)))
+
+    def send_bulk(self, payload):
+        raise NotImplementedError()
+
+
+def send_to_elasticsearch(path, es_getter, name, data=None, retries=MAX_RETRIES,
+        except_on_failure=False, update=False, delete=False):
     """
-    Elasticsearch handler
+    More fault tolerant es.put method
+    """
+    data = data if data is not None else {}
+    current_tries = 0
+    while current_tries < retries:
+        try:
+            if delete:
+                res = es_getter().delete(path=path)
+            elif update:
+                res = es_getter().post("%s/_update" % path, data={"doc": data})
+            else:
+                res = es_getter().put(path, data=data)
+            break
+        except ConnectionError, ex:
+            current_tries += 1
+            pillow_logging.error("[%s] put_robust error %s attempt %d/%d" % (
+                name, ex, current_tries, retries))
+            time.sleep(math.pow(RETRY_INTERVAL, current_tries))
+
+            if current_tries == retries:
+                message = "[%s] Max retry error on %s" % (name, path)
+                if except_on_failure:
+                    raise PillowtopIndexingError(message)
+                else:
+                    pillow_logging.error(message)
+                res = {}
+
+    if res.get('status', 0) == 400:
+        error_message = "Pillowtop put_robust error [%s]:\n%s\n\tpath: %s\n\t%s" % (
+            name,
+            res.get('error', "No error message"),
+            path,
+            data.keys())
+
+        if except_on_failure:
+            raise PillowtopIndexingError(error_message)
+        else:
+            pillow_logging.error(error_message)
+    return res
+
+
+class AliasedElasticPillow(BulkPillow):
+    """
+    This pillow class defines it as being alias-able. That is, when you query it, you use an
+    Alias to access it.
+
+    This could be for varying reasons -  to make an index by certain metadata into its own index
+    for performance/separation reasons
+
+    Or, for our initial use case, needing to version/update the index mappings on the fly with
+    minimal disruption.
+
+
+    The workflow for altering an AliasedElasticPillow is that if you know you made a change, the
+    pillow will create a new Index with a new md5sum as its suffix. Once it's finished indexing,
+    you will need to flip the alias over to it.
     """
     es_host = ""
     es_port = ""
     es_index = ""
     es_type = ""
+    es_alias = ''
+    es_index_prefix = ''
+    seen_types = {}
     es_meta = {}
     es_timeout = 3  # in seconds
     bulk = False
@@ -352,16 +559,22 @@ class ElasticPillow(BasicPillow):
         """
         create_index if the index doesn't exist on the ES cluster
         """
-        super(ElasticPillow, self).__init__(**kwargs)
+        super(AliasedElasticPillow, self).__init__(**kwargs)
         self.online = online
         index_exists = self.index_exists()
         if create_index and not index_exists:
             self.create_index()
+        if self.online:
+            self.seen_types = self.get_index_mapping()
+            pillow_logging.info("Pillowtop [%s] Retrieved mapping from ES" % self.get_name())
+        else:
+            pillow_logging.info("Pillowtop [%s] Started with no mapping from server in memory testing mode" % self.get_name())
+            self.seen_types = {}
 
     @autoretry_connection()
     def index_exists(self):
         if not self.online:
-            #If offline, just say the index is there and proceed along
+            # If offline, just say the index is there and proceed along
             return True
 
         es = self.get_es()
@@ -416,19 +629,24 @@ class ElasticPillow(BasicPillow):
         self.send_robust(self.es_index, data=self.es_meta)
         self.set_index_normal_settings()
 
+    def refresh_index(self):
+        self.get_es().post("%s/_refresh" % self.es_index)
+
     def change_trigger(self, changes_dict):
+        id = changes_dict['id']
         if changes_dict.get('deleted', False):
             try:
-                if self.get_es().head(path=self.get_doc_path(changes_dict['id'])):
-                    self.get_es().delete(path=self.get_doc_path(changes_dict['id']))
+                if self.get_es().head(path=self.get_doc_path(id)):
+                    self.get_es().delete(path=self.get_doc_path(id))
             except Exception, ex:
-                pillow_logging.error("ElasticPillow: error deleting route %s - ignoring: %s" % \
-                              (self.get_doc_path(changes_dict['id']), ex))
+                pillow_logging.error(
+                    "ElasticPillow: error deleting route %s - ignoring: %s" % (
+                        self.get_doc_path(changes_dict['id']),
+                        ex,
+                    )
+                )
             return None
-        if 'doc' in changes_dict:
-            return changes_dict['doc']
-        else:
-            return self.couch_db.open_doc(changes_dict['id'])
+        return super(AliasedElasticPillow, self).change_trigger(changes_dict)
 
     @autoretry_connection()
     def doc_exists(self, doc_id):
@@ -441,158 +659,77 @@ class ElasticPillow(BasicPillow):
         head_result = es.head(doc_path)
         return head_result
 
-    def bulk_builder(self, changes):
-        """
-        Generator function for bulk changes - note each individual change item goes through the pillowtop pathway individually
-        when loading the bulk item, and short of change_transport, it's identical. It would be slightly more efficient if the couch
-        load could be done in bulk for the actual documents, but it's not quite possible without gutting the existing pillowtop API
-        http://www.elasticsearch.org/guide/reference/api/bulk.html
-        bulk loader follows the following:
-        { "index" : { "_index" : "test", "_type" : "type1", "_id" : "1" } }\n
-        { "field1" : "value1" }\n
-        """
-        for change in changes:
-            try:
-                t = self.change_trigger(change)
-                if t is not None:
-                    tr = self.change_transform(t)
-                    if tr is not None:
-                        self.change_transport(tr)
+    def send_robust(self, path, data=None, retries=MAX_RETRIES,
+            except_on_failure=False, update=False):
+        return send_to_elasticsearch(
+            path=path,
+            es_getter=self.get_es,
+            name=self.get_name(),
+            data=data,
+            retries=retries,
+            except_on_failure=except_on_failure,
+            update=update,
+        )
 
-                        yield {
-                            "index": {"_index": self.es_index, "_type": self.es_type,
-                                      "_id": tr['_id']}}
-                        yield tr
-            except Exception, ex:
-                pillow_logging.error("[%s] Error on change: %s, %s" % (self.get_name(), change['id'], ex))
-
-    def process_bulk(self, changes):
-        self.allow_updates = False
-        self.bulk = True
-        es = self.get_es()
-        bstart = datetime.utcnow()
-        bulk_payload = '\n'.join(map(simplejson.dumps, self.bulk_builder(changes))) + "\n"
-        pillow_logging.info(
-            "%s,prepare_bulk,%s" % (self.get_name(), str(ms_from_timedelta(datetime.utcnow() - bstart) / 1000.0)))
-        send_start = datetime.utcnow()
-        es.post('_bulk', data=bulk_payload)
-        pillow_logging.info(
-            "%s,send_bulk,%s" % (self.get_name(), str(ms_from_timedelta(datetime.utcnow() - send_start) / 1000.0)))
-
-    def processor(self, change, do_set_checkpoint=True):
-        """
-        Parent processor for a pillow class - this should not be overridden.
-        This workflow is made for the situation where 1 change yields 1 transport/transaction
-        """
-        self.changes_seen += 1
-        if self.changes_seen % CHECKPOINT_FREQUENCY == 0 and do_set_checkpoint:
-            self.set_checkpoint(change)
-
-        self.process_change(change)
-
-    def send_robust(self, path, data={}, retries=MAX_RETRIES, except_on_failure=False, update=False):
-        """
-        More fault tolerant es.put method
-        """
-        current_tries = 0
-        while current_tries < retries:
-            try:
-                if update:
-                    res = self.get_es().post("%s/_update" % path, data={"doc": data})
-                else:
-                    res = self.get_es().put(path, data=data)
-                break
-            except ConnectionError, ex:
-                current_tries += 1
-                pillow_logging.error("[%s] put_robust error %s attempt %d/%d" % (
-                    self.get_name(), ex, current_tries, retries))
-                time.sleep(math.pow(RETRY_INTERVAL, current_tries))
-
-                if current_tries == retries:
-                    message = "[%s] Max retry error on %s" % (self.get_name(), path)
-                    if except_on_failure:
-                        raise PillowtopIndexingError(message)
-                    else:
-                        pillow_logging.error(message)
-                    res = {}
-
-        if res.get('status', 0) == 400:
-            error_message = "Pillowtop put_robust error [%s]:\n%s\n\tpath: %s\n\t%s" % (
-                self.get_name(),
-                res.get('error', "No error message"),
-                path,
-                data.keys())
-
-            if except_on_failure:
-                raise PillowtopIndexingError(error_message)
-            else:
-                pillow_logging.error(error_message)
-        return res
-
-    @autoretry_connection()
     def change_transport(self, doc_dict):
         """
-        Default elastic pillow for a given doc to a type.
+        Override the elastic transport to go to the index + the type being a string between the
+        domain and case type
         """
-        if self.bulk:
-            return
+        try:
+            if not self.type_exists(doc_dict):
+                # if type is never seen, apply mapping for said type
+                type_mapping = self.get_mapping_from_type(doc_dict)
+                # update metadata
+                type_mapping[self.get_type_string(doc_dict)]['_meta'][
+                    'created'] = datetime.isoformat(datetime.utcnow())
+                mapping_res = self.set_mapping(self.get_type_string(doc_dict), type_mapping)
+                if mapping_res.get('ok', False) and mapping_res.get('acknowledged', False):
+                    # API confirms OK, trust it.
+                    pillow_logging.info(
+                        "Mapping set: [%s] %s" % (self.get_type_string(doc_dict), mapping_res))
+                    # manually update in memory dict
+                    self.seen_types[self.get_type_string(doc_dict)] = {}
 
+            if not self.bulk:
+                doc_path = self.get_doc_path_typed(doc_dict)
+
+                doc_exists = self.doc_exists(doc_dict)
+
+                if self.allow_updates:
+                    can_put = True
+                else:
+                    can_put = not doc_exists
+
+                if can_put and not self.bulk:
+                    res = self.send_robust(doc_path, data=doc_dict, update=doc_exists)
+                    return res
+        except Exception, ex:
+            tb = traceback.format_exc()
+            pillow_logging.error(
+                "PillowTop [%(pillow_name)s]: Aliased Elastic Pillow transport change data doc_id: %(doc_id)s to elasticsearch error: %(error)s\ntraceback: %(tb)s\n" %
+                {
+                    "pillow_name": self.get_name(),
+                    "doc_id": doc_dict['_id'],
+                    "error": ex,
+                    "tb": tb
+                }
+            )
+            return None
+
+    def send_bulk(self, payload):
         es = self.get_es()
-        doc_path = self.get_doc_path(doc_dict['_id'])
-
-        doc_exists = self.doc_exists(doc_dict['_id'])
-
-        if self.allow_updates:
-            can_put = True
-        else:
-            can_put = not doc_exists
-
-        if can_put:
-            res = self.send_robust(doc_path, data=doc_dict, update=doc_exists)
-            return res
-
-
-class AliasedElasticPillow(ElasticPillow):
-    """
-    This pillow class defines it as being alias-able. That is, when you query it, you use an
-    Alias to access it.
-
-    This could be for varying reasons -  to make an index by certain metadata into its own index
-    for performance/separation reasons
-
-    Or, for our initial use case, needing to version/update the index mappings on the fly with
-    minimal disruption.
-
-
-    The workflow for altering an AliasedElasticPillow is that if you know you made a change, the
-    pillow will create a new Index with a new md5sum as its suffix. Once it's finished indexing,
-    you will need to flip the alias over to it.
-    """
-    es_alias = ''
-    es_index_prefix = ''
-    seen_types = {}
-    es_index = ""
-
-    def __init__(self, **kwargs):
-        super(AliasedElasticPillow, self).__init__(**kwargs)
-        if self.online:
-            self.seen_types = self.get_index_mapping()
-            pillow_logging.info("Pillowtop [%s] Retrieved mapping from ES" % self.get_name())
-        else:
-            pillow_logging.info("Pillowtop [%s] Started with no mapping from server in memory testing mode" % self.get_name())
-            self.seen_types = {}
-
+        es.post('_bulk', data=payload)
 
     def check_alias(self):
         """
         Naive means to verify the alias of the current pillow iteration is matched.
-        If we go fancier with routing and multi-index aliases due to index splitting, this
-        will need to be revisited.
         """
         es = self.get_es()
         aliased_indexes = es[self.es_alias].get('_aliases')
         return aliased_indexes.keys()
 
+    # todo: remove from class - move to the ptop_es_manage command
     def assume_alias(self):
         """
         For this instance, have the index that represents this index receive the alias itself.
@@ -620,7 +757,6 @@ class AliasedElasticPillow(ElasticPillow):
     def calc_mapping_hash(self, mapping):
         return hashlib.md5(simplejson.dumps(mapping, sort_keys=True)).hexdigest()
 
-
     def calc_meta(self):
         raise NotImplementedError("Need to implement your own meta calculator")
 
@@ -633,18 +769,25 @@ class AliasedElasticPillow(ElasticPillow):
         """
         for change in changes:
             try:
-                t = self.change_trigger(change)
-                if t is not None:
-                    tr = self.change_transform(t)
-                    if tr is not None:
-                        self.change_transport(tr)
-
-                        yield {"index": {"_index": self.es_index, "_type": self.get_type_string(tr),
-                                         "_id": tr['_id']}}
-                        yield tr
+                with lock_manager(self.change_trigger(change)) as t:
+                    if t is not None:
+                        tr = self.change_transform(t)
+                        if tr is not None:
+                            self.change_transport(tr)
+                            yield {
+                                "index": {
+                                    "_index": self.es_index,
+                                    "_type": self.get_type_string(tr),
+                                    "_id": tr['_id']
+                                }
+                            }
+                            yield tr
             except Exception, ex:
-                pillow_logging.error("Error on change: %s, %s" % (change['id'], ex))
+                pillow_logging.error(
+                    "Error on change: %s, %s" % (change['id'], ex)
+                )
 
+    # todo: make private and remove extra agruments that are never called
     def type_exists(self, doc_dict, server=False):
         """
         Verify whether the server has indexed this type
@@ -683,7 +826,7 @@ class AliasedElasticPillow(ElasticPillow):
     @autoretry_connection()
     def doc_exists(self, doc_dict):
         """
-        Overrided based upon the doc type
+        Overridden based upon the doc type
         """
         es = self.get_es()
         head_result = es.head(self.get_doc_path_typed(doc_dict))
@@ -701,50 +844,6 @@ class AliasedElasticPillow(ElasticPillow):
     def get_mapping_from_type(self, doc_dict):
         raise NotImplementedError("This must be implemented in this subclass!")
 
-    def change_transport(self, doc_dict):
-        """
-        Override the elastic transport to go to the index + the type being a string between the
-        domain and case type
-        """
-        try:
-            if not self.type_exists(doc_dict):
-                #if type is never seen, apply mapping for said type
-                type_mapping = self.get_mapping_from_type(doc_dict)
-                #update metadata
-                type_mapping[self.get_type_string(doc_dict)]['_meta'][
-                    'created'] = datetime.isoformat(datetime.utcnow())
-                mapping_res = self.set_mapping(self.get_type_string(doc_dict), type_mapping)
-                if mapping_res.get('ok', False) and mapping_res.get('acknowledged', False):
-                    #API confirms OK, trust it.
-                    pillow_logging.info(
-                        "Mapping set: [%s] %s" % (self.get_type_string(doc_dict), mapping_res))
-                    #manually update in memory dict
-                    self.seen_types[self.get_type_string(doc_dict)] = {}
-
-            if not self.bulk:
-                doc_path = self.get_doc_path_typed(doc_dict)
-
-                doc_exists = self.doc_exists(doc_dict)
-
-                if self.allow_updates:
-                    can_put = True
-                else:
-                    can_put = not doc_exists
-
-                if can_put and not self.bulk:
-                    res = self.send_robust(doc_path, data=doc_dict, update=doc_exists)
-                    return res
-        except Exception, ex:
-            tb = traceback.format_exc()
-            pillow_logging.error("PillowTop [%(pillow_name)s]: Aliased Elastic Pillow transport change data doc_id: %(doc_id)s to elasticsearch error: %(error)s\ntraceback: %(tb)s\n" %
-                          {
-                              "pillow_name": self.get_name(),
-                              "doc_id": doc_dict['_id'],
-                              "error": ex,
-                              "tb": tb
-                          })
-            return None
-
 
 class NetworkPillow(BasicPillow):
     """
@@ -789,3 +888,49 @@ class LogstashMonitoringPillow(NetworkPillow):
             return {}
         else:
             return super(NetworkPillow, self).processor(change)
+
+
+def retry_on_connection_failure(fn):
+    @wraps(fn)
+    def _inner(*args, **kwargs):
+        retry = kwargs.pop('retry', True)
+        try:
+            fn(*args, **kwargs)
+        except db.utils.DatabaseError:
+            # we have to do this manually to avoid issues with
+            # open transactions and already closed connections
+            db.transaction.rollback()
+            # re raise the exception for additional error handling
+            raise
+        except InterfaceError:
+            # force closing the connection to prevent Django from trying to reuse it.
+            # http://www.tryolabs.com/Blog/2014/02/12/long-time-running-process-and-django-orm/
+            db.connection.close()
+            if retry:
+                _inner(retry=False, *args, **kwargs)
+            else:
+                # re raise the exception for additional error handling
+                raise
+
+    return _inner
+
+
+class SQLPillowMixIn(object):
+
+    def change_trigger(self, changes_dict):
+        if changes_dict.get('deleted', False):
+            self.change_transport({'_id': changes_dict['id']}, delete=True)
+            return None
+        return super(SQLPillowMixIn, self).change_trigger(changes_dict)
+
+    @db.transaction.commit_on_success
+    @retry_on_connection_failure
+    def change_transport(self, doc_dict, delete=False):
+        self.process_sql(doc_dict, delete)
+
+    def process_sql(self, doc_dict, delete=False):
+        pass
+
+
+class SQLPillow(SQLPillowMixIn, BasicPillow):
+    pass
